"""
alignment preprocessing
"""

# BUILT-INS
import re
import multiprocessing

# varVAMP
from varvamp.scripts import config

# LIBS
from Bio import AlignIO
from Bio.Seq import Seq


def read_alignment(alignment_path):
    """
    read alignment with AlignIO and
    convert to list of lists
    """
    alignment_list = []

    for sequence in AlignIO.read(alignment_path, "fasta"):
        alignment_list.append([sequence.id, str(sequence.seq)])

    return alignment_list


def preprocess(alignment_path):
    """
    force nucleotides to lower and
    back transcribe if its RNA
    """

    preprocessed_alignment = []
    # read alignment
    alignment = read_alignment(alignment_path)

    for sequence in alignment:
        seq = Seq(sequence[1])
        seq = seq.lower()
        if "u" in seq:
            seq = seq.back_transcribe()
        preprocessed_alignment.append([sequence[0], str(seq)])

    return preprocessed_alignment


def find_internal_gaps(unique_gaps, gap):
    """
    find all unique gaps that
    lie within the current gap
    """
    overlapping_gaps = []

    if gap[1] - gap[0] == 0:
        # if the gap length = 1 there are
        # no overlapping gaps
        overlapping_gaps = [gap]
    else:
        # for each unique gap check if the intersection with the
        # gap is the same as the unique gap -> internal gap of
        # the current gap
        for unique_gap in unique_gaps:
            unique_set = set(range(unique_gap[0], unique_gap[1]))
            current_range = range(gap[0], gap[1])
            intersection = unique_set.intersection(current_range)
            if not intersection:
                continue
            if min(intersection) == unique_gap[0] and max(intersection) + 1 == unique_gap[1]:
                overlapping_gaps.append(unique_gap)

    return overlapping_gaps


def find_overlapping_gaps_worker(gap_list, unique_gaps):
    """
    Worker function to find overlapping gaps and count their occurrences.
    """
    gap_dict_part = {}
    for gap in gap_list:
        overlapping_gaps = find_internal_gaps(unique_gaps, gap)
        for overlapping_gap in overlapping_gaps:
            if overlapping_gap in gap_dict_part:
                gap_dict_part[overlapping_gap] += 1
            else:
                gap_dict_part[overlapping_gap] = 1
    return gap_dict_part


def create_gap_dictionary(unique_gaps, all_gaps, n_threads):
    """
    Creates a dictionary with all gap counts.
    Counts also all overlapping gaps per gap.
    Uses multiprocessing for parallelization.
    """

    with multiprocessing.Pool(processes=n_threads) as pool:
        results = pool.starmap(find_overlapping_gaps_worker, [(gap_list, unique_gaps) for gap_list in all_gaps])

    gap_dict = {}
    for gap_dict_part in results:
        for gap, count in gap_dict_part.items():
            if gap in gap_dict:
                gap_dict[gap] += count
            else:
                gap_dict[gap] = count

    return gap_dict


def find_gaps_to_mask(gap_dict, cutoff):
    """
    filters gaps for their freq cutoff.
    condenses final gaps if there is
    an overlap.
    """
    gaps_to_mask = []
    potential_gaps = []
    opened_region = []

    # check for each region if it is covered
    # by enough sequences
    for gap in gap_dict:
        if gap_dict[gap] > cutoff:
            potential_gaps.append(gap)

    # sort by start and stop
    potential_gaps = sorted(potential_gaps)

    # get the min and max of overlapping gaps
    for i, region in enumerate(potential_gaps):
        region = list(region)
        if opened_region:
            # write the opened region if the start of the current region
            # > opened_region[stop] and the last still opened region
            if region[0] > opened_region[1] or i == len(potential_gaps) - 1:
                gaps_to_mask.append(opened_region)
                opened_region = region
            else:
                # 1 case: same start and further stop -> new stop
                if region[0] == opened_region[0]:
                    opened_region[1] = region[1]
                # 2 case: further start and further stop -> new stop
                if region[0] > opened_region[0] and region[1] > opened_region[1]:
                    opened_region[1] = region[1]
        else:
            opened_region = region

    return gaps_to_mask


def clean_gaps(alignment, gaps_to_mask):
    """
    clean an alignment of large common deletions.
    """
    cleaned_alignment = []

    for sequence in alignment:
        start = 0
        masked_seq = str()
        for region in gaps_to_mask:
            # check if it is three bases or more and mask with 2 Ns
            if region[1] - region[0] >= config.QAMPLICON_DEL_CUTOFF:
                mask = "NN"
            # or mask with one N (small deletion)
            else:
                mask = "N"
            stop = region[0]
            masked_seq_temp = sequence[1][start:stop]
            # check if the deletion is at the start
            if start == 0 and len(masked_seq_temp) == 0:
                masked_seq = mask
            # check if deletion is not at start
            elif start == 0 and len(masked_seq_temp) != 0:
                masked_seq = masked_seq_temp
            # else we are in the middle of the alignment
            else:
                masked_seq = masked_seq + mask + masked_seq_temp
            start = region[1] + 1
        if max(gaps_to_mask)[1] < len(sequence[1]) - 1:
            # append the last seq if no gap is at
            # the end of the sequence
            start = max(gaps_to_mask)[1]
            stop = len(sequence[1]) - 1
            masked_seq_temp = sequence[1][start:stop]
            masked_seq = masked_seq + mask + masked_seq_temp
        else:
            # append the mask to the end of the seq
            masked_seq = masked_seq + mask

        cleaned_alignment.append([sequence[0], masked_seq])

    return cleaned_alignment


def process_alignment(preprocessed_alignment, threshold, n_threads):
    """
    proprocesses alignment and cleans gaps
    """
    all_gaps = []

    gap_cutoff = len(preprocessed_alignment) * (1 - threshold)
    for seq in preprocessed_alignment:
        # find all gaps for all sequences with regular expression -{min}
        all_gaps.append(
            [(gap.start(0), gap.end(0) - 1) for gap in re.finditer(
                "-{1,}", seq[1])]
        )
    unique_gaps = list(set(gaps for gap_list in all_gaps for gaps in gap_list))

    if unique_gaps:
        gap_dic = create_gap_dictionary(unique_gaps, all_gaps, n_threads)
        gaps_to_mask = find_gaps_to_mask(gap_dic, gap_cutoff)
        if gaps_to_mask:
            alignment_cleaned = clean_gaps(
                preprocessed_alignment, gaps_to_mask
            )
        else:
            alignment_cleaned = preprocessed_alignment
    else:
        gaps_to_mask = []
        alignment_cleaned = preprocessed_alignment

    return alignment_cleaned, gaps_to_mask


def calculate_total_masked_gaps(gaps_to_mask):
    """
    calculates the cummulative length of gaps
    that were masked.
    """
    if gaps_to_mask:
        sum_gaps = 0
        for region in gaps_to_mask:
            sum_gaps += region[1] - region[0] + 1
        return sum_gaps
    else:
        return 0



"""
qPCR probe and amplicon design
"""
import re

# LIBS
import seqfold
import itertools
import multiprocessing

# varVAMP
from varvamp.scripts import config
from varvamp.scripts import primers
from varvamp.scripts import reporting


def choose_probe_direction(seq):
    """
    choose the direction of the probe so
    that the probe always has more c than g
    """
    c_count = seq.count("c")
    g_count = seq.count("g")
    direction = "-+"
    if c_count < g_count:
        direction = "-"
    if c_count > g_count:
        direction = "+"

    return direction


def ambiguous_ends(amb_seq):
    """
    determine if kmers have ambiguous ends
    """
    if amb_seq[:1] in config.NUCS and amb_seq[-1] in config.NUCS:
        return False
    else:
        return True


def filter_probe_direction_dependent(seq):
    """
    filter probes for their direction
    """
    return (
            (seq[:1] != "g")  # no 5' g as this leads to quenching
            and (primers.calc_hairpin(seq).tm <= config.PRIMER_HAIRPIN)
            and (config.QPROBE_GC_END[0] <= primers.calc_end_gc(seq) <= config.QPROBE_GC_END[1])
    )


def get_qpcr_probes(kmers, ambiguous_consensus, alignment_cleaned):
    """
    find potential qPCR probes
    """
    probe_candidates = {}
    probe_idx = 0

    for kmer in kmers:
        # filter probe for base params
        if not primers.filter_kmer_direction_independent(kmer[0], config.QPROBE_TMP, config.QPROBE_GC_RANGE,
                                                         config.QPROBE_SIZES):
            continue
        # do not allow ambiguous chars at both ends
        if ambiguous_ends(ambiguous_consensus[kmer[1]:kmer[2]]):
            continue
        # calc penalties analogous to primer search
        base_penalty = primers.calc_base_penalty(kmer[0], config.QPROBE_TMP, config.QPROBE_GC_RANGE,
                                                 config.QPROBE_SIZES)
        per_base_mismatches = primers.calc_per_base_mismatches(
            kmer,
            alignment_cleaned,
            ambiguous_consensus
        )
        permutation_penalty = primers.calc_permutation_penalty(
            ambiguous_consensus[kmer[1]:kmer[2]]
        )
        # determine the direction with more cytosine or set both if 50 %
        direction = choose_probe_direction(kmer[0])
        # create probe dictionary
        if "+" in direction:
            if filter_probe_direction_dependent(kmer[0]):
                probe_name = f"PROBE_{probe_idx}_FW"
                three_prime_penalty = primers.calc_3_prime_penalty("+", per_base_mismatches)
                probe_candidates[probe_name] = [kmer[0], kmer[1], kmer[2],
                                                base_penalty + permutation_penalty + three_prime_penalty,
                                                per_base_mismatches, direction]
                probe_idx += 1
        if "-" in direction:
            if filter_probe_direction_dependent(primers.rev_complement(kmer[0])):
                probe_name = f"PROBE_{probe_idx}_RV"
                three_prime_penalty = primers.calc_3_prime_penalty("-", per_base_mismatches)
                probe_candidates[probe_name] = [primers.rev_complement(kmer[0]), kmer[1], kmer[2],
                                                base_penalty + permutation_penalty + three_prime_penalty,
                                                per_base_mismatches, direction]
                probe_idx += 1
    # sort by penalty
    probe_candidates = dict(sorted(probe_candidates.items(), key=lambda x: x[1][3]))

    return probe_candidates


def flanking_primer_subset(primer_list, primer_direction, probe):
    """
    subset for primers flanking the probe and sort by penalty
    """
    subset = []

    if primer_direction == "+":
        window_start = probe[1] - config.QAMPLICON_LENGTH[1] + len(probe[0])
        window_stop = probe[1]
    else:
        window_start = probe[2]
        window_stop = probe[2] + config.QAMPLICON_LENGTH[1] - len(probe[0])
    for primer in primer_list:
        if window_start < primer[1] and primer[2] < window_stop:
            subset.append(primer)

    # sort by penalty and start if same penalty
    subset.sort(key=lambda x: (x[3], x[1]))

    return subset


def hardfilter_amplicon(majority_consensus, left_primer, right_primer):
    """
    hardfilter possible amplicon for length, gc and if there are Ns
    (indicating deletions in amp) or n (no folding analysis possible)
    """
    amplicon_length = right_primer[2] - left_primer[1]
    amplicon_seq = majority_consensus[left_primer[1]:right_primer[2]]
    # check for the right amplicon length
    return (
            (config.QAMPLICON_LENGTH[0] <= amplicon_length <= config.QAMPLICON_LENGTH[1])
            and (config.QAMPLICON_GC[0] <= primers.calc_gc(amplicon_seq) <= config.QAMPLICON_GC[1])
            and "NN" not in amplicon_seq
    )


def check_end_overlap(dimer_result):
    """
    checks if two oligos overlap at their ends (pretty rare)
    Example:
        xxxxxxxxtagc-------
        --------atcgxxxxxxx
    """
    if dimer_result.structure_found:
        # clean structure
        structure = [x[4:] for x in dimer_result.ascii_structure_lines]
        # calc overlap and the cumulative len of the oligos
        overlap = len(structure[1].replace(" ", ""))
        nt_count = len(re.findall("[ATCG]", "".join(structure)))
        # check for overlaps at the ends and the min overlap (allows for some amount of mismatches)
        if overlap > config.END_OVERLAP and nt_count <= len(structure[0]) + overlap + 1 and "  " not in structure[1].lstrip(" "):
            return True

    return False


def forms_dimer_or_overhangs(right_primer, left_primer, probe, ambiguous_consensus):
    """
    checks if combinations of primers/probe form dimers or overhangs
    """

    forms_structure = False

    # first check if there are dimers between the two flanking primers
    if primers.calc_dimer(left_primer[0], right_primer[0]).tm > config.PRIMER_MAX_DIMER_TMP:
        return True
    # for the probe check all permutations and possible overhangs to ensure
    # that none of the primers could cause unspecific probe binding.
    # first get all permutations
    probe_per = reporting.get_permutations(ambiguous_consensus[probe[1]:probe[2]])
    left_per = reporting.get_permutations(ambiguous_consensus[left_primer[1]:left_primer[2]])
    right_per = reporting.get_permutations(ambiguous_consensus[right_primer[1]:right_primer[2]])
    # then check all permutations
    for combination in [(probe_per, left_per), (probe_per, right_per)]:
        for oligo1 in combination[0]:
            for oligo2 in combination[1]:
                dimer_result = primers.calc_dimer(oligo1, oligo2, structure=True)
                if dimer_result.tm >= config.PRIMER_MAX_DIMER_TMP or check_end_overlap(dimer_result):
                    forms_structure = True
                    break
            # break all loops because we found an unwanted structure in one of the permutations
            # (either dimer formation or a too long overlap at the ends of the primer)
            if forms_structure:
                break
        if forms_structure:
            break

    return forms_structure


def assess_amplicons(left_subset, right_subset, qpcr_probes, probe, majority_consensus, ambiguous_consensus):
    """
    assess if a potential amplicon is a qPCR scheme for a specific probe and return the one with the lowest penalty
    """

    primer_combinations = ()
    amplicon_found = False

    # consider a combination of flanking primers if ...
    for left_primer in left_subset:
        for right_primer in right_subset:
            # ... the amplicon is large enough and is in gc range, ...
            if not hardfilter_amplicon(majority_consensus, left_primer, right_primer):
                continue
            # ... the probe is close enough to the primer on the same strand
            if "FW" in probe:
                if not qpcr_probes[probe][1] in range(
                        left_primer[2] + config.QPROBE_DISTANCE[0],
                        left_primer[2] + config.QPROBE_DISTANCE[1] + 1
                ):
                    continue
            elif "RV" in probe:
                if not right_primer[1] in range(
                        qpcr_probes[probe][2] + config.QPROBE_DISTANCE[0],
                        qpcr_probes[probe][2] + config.QPROBE_DISTANCE[1] + 1

                ):
                    continue
            # ... the primer temps do not differ too much, ...
            primer_temps = (primers.calc_temp(right_primer[0]), primers.calc_temp(left_primer[0]))
            if abs(primer_temps[0] - primer_temps[1]) > config.QPRIMER_DIFF:
                continue
            # ... the probe has a higher temp than the primers and ...
            probe_temp = primers.calc_temp(qpcr_probes[probe][0])
            if not all(
                    [config.QPROBE_TEMP_DIFF[0] <= probe_temp - x <= config.QPROBE_TEMP_DIFF[1] for x in primer_temps]):
                continue
            # .... all combination of oligos do not form dimers or overhangs.
            if forms_dimer_or_overhangs(right_primer, left_primer, qpcr_probes[probe], ambiguous_consensus):
                continue
            # append to list and break as this is the primer combi
            # with the lowest penalty (primers are sorted by penalty)
            amplicon_found = True
            break
        # break also the outer loop
        if amplicon_found:
            primer_combinations = (left_primer, right_primer)
            break

    return primer_combinations


def find_qcr_schemes(qpcr_probes, left_primer_candidates, right_primer_candidates, majority_consensus,
                     ambiguous_consensus):
    """
    this finds the final qPCR schemes. it slices for primers flanking a probe and
    test all left/right combinations whether they are potential amplicons. as primers
    are sorted by penalty, only the very first match is considered as this has the
    lowest penalty. however, probes are overlapping and there is a high chance that
    left and right primers are found multiple times. to consider only one primer-probe
    combination the probes are also sorted by penalty. therefore, if a primer
    combination has been found already the optimal probe was already selected and
    there is no need to consider this primer probe combination.
    """

    qpcr_scheme_candidates = {}
    found_amplicons = []
    amplicon_nr = -1

    for probe in qpcr_probes:
        left_subset = flanking_primer_subset(left_primer_candidates, "+", qpcr_probes[probe])
        right_subset = flanking_primer_subset(right_primer_candidates, "-", qpcr_probes[probe])
        # consider if there are primers flanking the probe ...
        if not left_subset or not right_subset:
            continue
        primer_combination = assess_amplicons(left_subset, right_subset, qpcr_probes, probe, majority_consensus,
                                              ambiguous_consensus)
        # ... a combi has been found, ...
        if not primer_combination:
            continue
        # ...and this combi is not already present for a probe with a better penalty.
        if primer_combination in found_amplicons:
            continue
        # populate the primer dictionary:
        amplicon_nr += 1
        found_amplicons.append(primer_combination)
        qpcr_scheme_candidates[f"AMPLICON_{amplicon_nr}"] = {
            "penalty": qpcr_probes[probe][3] + primer_combination[0][3] + primer_combination[1][3],
            "PROBE": qpcr_probes[probe],
            "LEFT": primer_combination[0],
            "RIGHT": primer_combination[1]
        }
    # and again sort by total penalty (left + right + probe)
    qpcr_scheme_candidates = dict(sorted(qpcr_scheme_candidates.items(), key=lambda x: x[1]["penalty"]))

    return qpcr_scheme_candidates


def process_single_amplicon_deltaG(amplicon, majority_consensus):
    """
    Process a single amplicon to test its deltaG and apply filtering.
    This function will be called concurrently by multiple threads.
    """
    name, data = amplicon
    start = data["LEFT"][1]
    stop = data["RIGHT"][2]
    seq = majority_consensus[start:stop]
    seq = seq.replace("N", "")
    seq = seq.replace("n", "")
    amp_positions = list(range(start, stop + 1))
    # check if the amplicon overlaps with an amplicon that was previously
    # found and had a high enough deltaG
    min_temp = min((primers.calc_temp(data["LEFT"][0]),
                    primers.calc_temp(data["RIGHT"][0])))
    # calculate deltaG at the minimal primer temp
    deltaG = seqfold.dg(seq, min_temp)

    return deltaG, amp_positions, name


def test_amplicon_deltaG_parallel(qpcr_schemes_candidates, majority_consensus, n_to_test, deltaG_cutoff, n_threads):
    """
    Test all amplicon deltaGs for the top n hits at the lowest primer temperature
    and filters if they fall below the cutoff. Multiple processes are used
    for processing amplicons in parallel.
    """
    final_schemes = {}
    passed_counter = 0  # counter for re-naming amplicons that passed deltaG cutoff
    amplicon_set = set()

    # Create a pool of processes to handle the concurrent processing
    with multiprocessing.Pool(processes=n_threads) as pool:
        # Create a list of the first n amplicon tuples for processing
        amplicons = itertools.islice(qpcr_schemes_candidates.items(), n_to_test)
        # process amplicons concurrently
        results = pool.starmap(process_single_amplicon_deltaG, [(amp, majority_consensus) for amp in amplicons])
        # Process the results
        for deltaG, amp_positions, amp_name in results:
            # check if the amplicon overlaps with an amplicon that was previously
            # found and had a high enough deltaG
            if any(x in amp_positions for x in amplicon_set):
                continue
            # and if this passes cutoff make a dict entry and do not allow further
            # amplicons in that region (they will have a lower penalty)
            if deltaG > deltaG_cutoff:
                new_name = f"QPCR_SCHEME_{passed_counter}"
                final_schemes[new_name] = qpcr_schemes_candidates[amp_name]
                final_schemes[new_name]["deltaG"] = deltaG
                amplicon_set.update(amp_positions)
                passed_counter += 1

    return final_schemes



"""
estimate varVAMP threshold and max n of ambiguous bases if none are given
"""

# varVAMP
from varvamp.scripts import config


def calculate_frequencies(preprocessed_alignment):
    """
    calculate the max nucleotide freq at each pos
    """
    all_freqs = []

    for i in range(0, len(preprocessed_alignment[0][1])):
        nuc_dict = {
            "a": 0,
            "c": 0,
            "t": 0,
            "g": 0,
        }
        for seq in preprocessed_alignment:
            if seq[1][i] in nuc_dict:
                nuc_dict[seq[1][i]] += 1
        highest_freq = max(nuc_dict.values())
        all_freqs.append(highest_freq/len(preprocessed_alignment))

    return all_freqs


def calculate_distances(all_freqs, threshold):
    """
    calc the distance for each  nuc freq to the prior
    nuc freq that fell below the cutoff
    """

    current_dis = 0
    previous_dis = 0
    all_dis = []

    for idx, freq in enumerate(all_freqs):
        if freq < threshold:
            current_dis = 0
        current_dis += 1
        if current_dis <= previous_dis or idx == len(all_freqs)-1:
            all_dis.append(previous_dis)

        previous_dis = current_dis

    return all_dis


def get_parameters(preprocessed_alignment, args, log_file):
    """
    give an estimate for number of ambiguous chars and/or threshold
    writes to log file
    """
    # set coverage to max
    coverage = 1
    # read in the alignment and calc freqs
    frequencies = calculate_frequencies(preprocessed_alignment)
    # if no args for both threshold and n_ambig are given
    # set the n_ambig to 2 and optimize threshold
    if args.threshold is None:
        args.threshold = 0.1
        if args.n_ambig is None:
            args.n_ambig = 2
        text = f"AUTOMATIC PARAMETER SELECTION\nvarVAMP estimates the threshold at {args.n_ambig} ambiguous bases"
        fixed = False
    # if threshold is given, optimize n_ambig (number of ambiguous chars)
    else:
        args.n_ambig = config.PRIMER_SIZES[0]
        text = f"varVAMP estimates the number of ambiguous bases at a threshold of {args.threshold}"
        fixed = True
    # write to log
    with open(log_file, 'a') as f:
        print(f"{text}\nto consider ~50% of the alignment for potential primers:\n\n-t\t-a\testimated coverage", file=f)

        # optimize until less than 50 % is covered
        while coverage >= 0.5 and args.threshold < 1:
            distances = calculate_distances(frequencies, args.threshold)
            # calculate the cummulative sum of the sum of n conseq. streches
            # that are together larger than the min primer length
            covered_pos = sum(
                [distances[x] for x in range(0, len(distances)) if sum(distances[x:x+args.n_ambig+1]) >= config.PRIMER_SIZES[0]]
            )
            # calculate coverage
            coverage = (covered_pos+1)/len(preprocessed_alignment[0][1])
            # change the non fixed param if threshold has not been reached
            if coverage >= 0.5:
                # write each iteration to log
                print(round(args.threshold, 2), args.n_ambig, round(coverage*100, 1), sep="\t", file=f)
                if fixed:
                    args.n_ambig -= 1
                else:
                    args.threshold += 0.01
            # or reset to the param of the prior iteration
            else:
                if fixed:
                    args.n_ambig += 1
                else:
                    args.threshold -= 0.01
                break
        print(f"Automatic parameter selection set -t {round(args.threshold, 2)} and -a {args.n_ambig}.", file=f)

    return round(args.threshold, 2), args.n_ambig



"""
varVAMP logging and raising errors
"""

# BUILT-INS
import sys
import os
import shutil
import datetime
import random
import statistics

# varVAMP
from varvamp.scripts import config
from varvamp import __version__


def create_dir_structure(dir):
    """
    create output folders and log file
    """
    cwd = os.getcwd()
    results_dir = os.path.join(cwd, dir)
    data_dir = os.path.join(results_dir, "data/")
    # create folders
    if not os.path.exists(results_dir):
        os.makedirs(results_dir)
    else:
        shutil.rmtree(results_dir)
        os.makedirs(results_dir)
    os.makedirs(data_dir)

    log_file = os.path.join(results_dir, "varvamp_log.txt")

    return results_dir, data_dir, log_file


def varvamp_progress(log_file, mode=None, start_time=None, progress=0, job="", progress_text=""):
    """
    progress bar, main progress logging and folder creation
    """
    barLength = 40
    block = int(round(barLength * progress))

    if progress == 0:
        print(
            "\nStarting \033[31m\033[1mvarVAMP ◥(ºwº)◤\033[0m primer design\n",
            flush=True
        )
        with open(log_file, 'w') as f:
            print(f"VARVAMP log \n\nMODE = {mode}\n", file=f)
    else:
        if progress == 1:
            stop_time = datetime.datetime.now() - start_time
            progress_text = f"all done \n\n\rvarVAMP {__version__} finished in {stop_time.total_seconds()} sec!\n{datetime.datetime.now()}"
            job = "Finalizing output."
        print(
            "\rJob:\t\t " + job + "\nProgress: \t [{0}] {1}%".format("█" * block + "-" * (barLength - block),
                                                                     progress * 100) + "\t" + progress_text,
            flush=True
        )
        with open(log_file, 'a') as f:
            print(
                f"\rJob:\t {job} \nResult:\t {progress_text}",
                file=f
            )


def raise_error(message, log_file, exit=False):
    """
    raises warnings or errors, writes to log
    """
    # print to log
    with open(log_file, 'a') as f:
        if exit:
            print(f"ERROR: {message}", file=f)
        else:
            print(f"WARNING: {message}", file=f)
    # print to console
    if exit:
        sys.exit(f"\n\033[31m\033[1mERROR:\033[0m {message}")
    else:
        print(f"\033[31m\033[1mWARNING:\033[0m {message}")


def raise_arg_errors(args, log_file):
    """
    checks arguments for non-valid input and raises warnings
    """
    # threshold error
    if args.threshold > 1 or args.threshold < 0:
        raise_error(
            "threshold can only be between 0-1",
            log_file,
            exit=True
        )
    if args.n_ambig < 0:
        raise_error(
            "number of ambiguous chars can not be negative",
            log_file,
            exit=True
        )
    if args.n_ambig > 4:
        raise_error(
            "high number of ambiguous nucleotides in primer leads to a high "
            "degeneracy. Consider reducing.",
            log_file
        )
    if args.database is not None:
        if args.threads < 1:
            raise_error(
                "number of threads cannot be smaller than 1.",
                log_file,
                exit=True
            )
    if args.mode in ("tiled", "single"):
        if args.opt_length > args.max_length:
            raise_error(
                "optimal length can not be higher than the maximum amplicon length.",
                log_file,
                exit=True
            )
        if args.opt_length < 0 or args.max_length < 0:
            raise_error(
                "amplicon lengths can not be negative.",
                log_file,
                exit=True
            )
    # SINGLE specific warnings
    if args.mode == "single":
        if args.report_n < 1:
            raise_error(
                "number of reported amplicons cannot be below 1.",
                log_file,
                exit=True
            )
    # TILED specific warnings
    if args.mode == "tiled":
        if args.opt_length < 200 or args.max_length < 200:
            raise_error(
                "your amplicon lengths might be to small. Consider increasing",
                log_file
            )
        if args.overlap < 0:
            raise_error(
                "overlap size can not be negative.",
                log_file,
                exit=True
            )
        if args.overlap < 50:
            raise_error(
                "small overlaps might hinder downstream analyses. Consider increasing.",
                log_file
            )
        if args.overlap > args.max_length / 2 - config.PRIMER_SIZES[1]:
            raise_error(
                "min overlap must be lower than half of your maximum length - maximum primer length. To achieve optimal results reduce it to at least half of your optimal length",
                log_file,
                exit=True
            )
        if args.overlap > args.opt_length:
            raise_error(
                "overlap can not be higher than your optimal length.",
                log_file,
                exit=True
            )
        if args.overlap > args.opt_length / 2:
            raise_error(
                "your intended overlap is higher than half of your optimal length. This reduces how well varvamps will find overlapping amplicons. Consider decreasing.",
                log_file
            )
    # QPCR specific warnings
    if args.mode == "qpcr":
        if args.pn_ambig < 0:
            raise_error(
                "number of ambiguous characters in the qPCR probe cannot be 0.",
                log_file,
                exit=True
            )
        if args.pn_ambig > args.n_ambig:
            raise_error(
                "the degeneracy of qPCR probes should not be higher than that of qPCR primers to retain specificity",
                log_file,
                exit=True
            )
        if args.test_n <= 0:
            raise_error(
                "if the number of tested qPCR amplicons is 0 or lower, no amplicons will be reported.",
                log_file,
                exit=True
            )

        if args.test_n > 200:
            raise_error(
                "checking the deltaG of amplicons is computationally intensive. setting this higher than 200 will likely take a bit to compute.",
                log_file
            )
        if args.pn_ambig > 2:
            raise_error(
                "setting the degeneracy of probes higher than 2 can result in variable probe design. Consider reducing!",
                log_file
            )
        if args.deltaG > 0:
            raise_error(
                "setting the deltaG cutoff larger than 0 is not recommended!",
                log_file
            )


def check_alignment_length(preprocessed_alignment, log_file):
    """
    checks the sequence lengths of the alignment and reports a warning
    if the length is larger or smaller than the mean +- 3std of all seqs
    and at least 1 % diff to mean
    """

    all_seq_length, all_names = [], []
    # count the length
    for seq in preprocessed_alignment:
        all_names.append(seq[0])
        all_seq_length.append(len(seq[1].strip("-")))
    # clac mean and std
    mean_len, mean_std = statistics.mean(all_seq_length), statistics.stdev(all_seq_length)
    # check for each seq if it is larger or smaller than the mean +-3std
    # and is at least 1 % smaller or larger than the alignment mean
    # (otherwise varvamp will report very small differences).
    smaller_warning, larger_warning = [], []
    for name, length in zip(all_names, all_seq_length):
        # consider variation of sequence lengths and check if it is at least 2 % deviation
        if length <= mean_len - 3 * mean_std and length <= mean_len - mean_len * 0.02:
            smaller_warning.append(f"{name} ({length} nt)\n")
        elif length >= mean_len + 3 * mean_std and length >= mean_len + mean_len * 0.02:
            larger_warning.append(f"{name} ({length} nt)\n")
    # raise warning for non-empty lists
    for warning, length_type in zip([larger_warning, smaller_warning], ["larger", "smaller"]):
        if not warning:
            continue
        warning = "".join(warning)
        raise_error(
            f"The following sequences are considerably {length_type} than the alignment mean ({round(mean_len)} nt) and might cause alignment trimming:\n{warning}",
            log_file,
            exit=False
        )


def confirm_config(args, log_file):
    """
    checks the config. raises error and warnings
    if nececarry. writes settings to log
    """
    error = False

    # check if all variables exists
    all_vars = [
        # arg independent TILED, SINGLE mode
        (
            "PRIMER_TMP",
            "PRIMER_GC_RANGE",
            "PRIMER_SIZES",
            "PRIMER_HAIRPIN",
            "PRIMER_MAX_POLYX",
            "PRIMER_MAX_DINUC_REPEATS",
            "PRIMER_GC_END",
            "PRIMER_MAX_DIMER_TMP",
            "PRIMER_MIN_3_WITHOUT_AMB",
            "PCR_MV_CONC",
            "PCR_DV_CONC",
            "PCR_DNTP_CONC",
            "PCR_DNA_CONC",
            "PRIMER_TM_PENALTY",
            "PRIMER_GC_PENALTY",
            "PRIMER_SIZE_PENALTY",
            "PRIMER_MAX_BASE_PENALTY",
            "PRIMER_3_PENALTY",
            "PRIMER_PERMUTATION_PENALTY"
        ),
        # arg independent QPCR mode
        (
            "QPROBE_TMP",
            "QPROBE_SIZES",
            "QPROBE_GC_RANGE",
            "QPROBE_GC_END",
            "QPRIMER_DIFF",
            "QPROBE_TEMP_DIFF",
            "QPROBE_DISTANCE",
            "QAMPLICON_LENGTH",
            "QAMPLICON_GC",
            "QAMPLICON_DEL_CUTOFF"
        ),
        (
            "BLAST_MAX_DIFF",
            "BLAST_SIZE_MULTI",
            "BLAST_PENALTY"
        )
    ]

    for tup in all_vars:
        for var in tup:
            if var not in vars(config):
                raise_error(
                    f"{var} does not exist in config!",
                    log_file
                )
                error = True
    # exit if variables are not defined
    if error:
        raise_error(
            "config is missing parameters. Look at the above warnings!",
            log_file,
            exit=True
        )
    # confirm tuples with 3 elements
    for tup_type, tup in [("primer temp", config.PRIMER_TMP), ("primer gc", config.PRIMER_GC_RANGE),
                          ("primer size", config.PRIMER_SIZES), ("probe temp", config.QPROBE_TMP),
                          ("probe size", config.QPROBE_SIZES), ("probe gc", config.QPROBE_GC_RANGE)]:
        if len(tup) != 3:
            raise_error(
                f"{tup_type} tuple has to have the form (min, max, opt)!",
                log_file
            )
            error = True
        if tup[0] > tup[1]:
            raise_error(
                f"min {tup_type} should not exceed max {tup_type}!",
                log_file
            )
            error = True
        if tup[0] > tup[2]:
            raise_error(
                f"min {tup_type} should not exceed opt {tup_type}!",
                log_file
            )
            error = True
        if tup[2] > tup[1]:
            raise_error(
                f"opt {tup_type} should not exeed max {tup_type}!",
                log_file
            )
            error = True
        if any(map(lambda var: var < 0, tup)):
            raise_error(
                f"{tup_type} can not contain negative values!",
                log_file
            )
            error = True
    # confirm tuples with two elements
    for tup_type, tup in [("primer gc at the 3'", config.PRIMER_GC_END),
                          ("probe and primer temp diff", config.QPROBE_TEMP_DIFF),
                          ("distance probe to primer", config.QPROBE_DISTANCE),
                          ("qpcr amplicon length", config.QAMPLICON_LENGTH), ("qpcr amplicon gc", config.QAMPLICON_GC),
                          ("probe gc at the 3'", config.PRIMER_GC_END)]:
        if len(tup) != 2:
            raise_error(
                f"{tup_type} tuple has to have the form (min, max)!",
                log_file
            )
            error = True
        if tup[0] > tup[1]:
            raise_error(
                f"min {tup_type} should not exceed max {tup_type}!",
                log_file
            )
            error = True
        if any(map(lambda var: var < 0, tup)):
            raise_error(
                f"{tup_type} can not contain negative values!",
                log_file
            )
            error = True
    # check single values that cannot be negative
    non_negative_var = [
        ("min number of 3 prime nucleotides without ambiguous nucleotides", config.PRIMER_MIN_3_WITHOUT_AMB),
        ("monovalent cation concentration", config.PCR_MV_CONC),
        ("divalent cation concentration", config.PCR_DV_CONC),
        ("dNTP concentration", config.PCR_DNTP_CONC),
        ("primer temperature penalty", config.PRIMER_TM_PENALTY),
        ("primer gc penalty", config.PRIMER_GC_PENALTY),
        ("primer size penalty", config.PRIMER_SIZE_PENALTY),
        ("max base penalty", config.PRIMER_MAX_BASE_PENALTY),
        ("primer permutation penalty", config.PRIMER_PERMUTATION_PENALTY),
        ("qpcr flanking primer difference", config.QPRIMER_DIFF),
        ("qpcr deletion size still considered for deltaG calculation", config.QAMPLICON_DEL_CUTOFF),
        ("maximum difference between primer and blast db", config.BLAST_MAX_DIFF),
        ("multiplier of the maximum length for non-specific amplicons", config.BLAST_SIZE_MULTI),
        ("blast penalty for off targets", config.BLAST_PENALTY)
    ]
    for var_type, var in non_negative_var:
        if var < 0:
            raise_error(
                f"{var_type} can not be negative!",
                log_file
            )
            error = True
    if any(map(lambda var: var < 0, config.PRIMER_3_PENALTY)):
        raise_error(
            "3' penalties can not be zero!",
            log_file
        )
        error = True
    # exit if variables are not properly defined
    if error:
        raise_error(
            "config has flaws. Look at the above warnings!",
            log_file,
            exit=True
        )
    # specific warnings
    if config.PRIMER_MAX_POLYX > 5:
        raise_error(
            "max polyx > 5 is not recommended.",
            log_file,
        )
    if config.PRIMER_MAX_DINUC_REPEATS > 5:
        raise_error(
            "max di-nucleotide repeats > 5 is not recommended.",
            log_file,
        )
    if config.PRIMER_HAIRPIN < 0:
        raise_error(
            "decreasing hairpin melting temp to negative values "
            "will influence successful primer search!",
            log_file
        )
    if config.PRIMER_MAX_DIMER_TMP < 0:
        raise_error(
            "there is no need to set max dimer melting temp below 0.",
            log_file
        )
    if config.PRIMER_MAX_BASE_PENALTY < 8:
        raise_error(
            "decreasing the base penalty will filter out more primers.",
            log_file
        )
    if config.PRIMER_GC_END[0] > 3 or config.QPROBE_GC_END[0] > 3:
        raise_error(
            "large GC clamps will results in too high 3'end stability",
            log_file
        )
    if config.PRIMER_GC_END[1] > 5 or config.QPROBE_GC_END[1] > 5:
        raise_error(
            "only the last 5 nucleotides of the 3' end are considered for GC 3'end calculation.",
            log_file
        )
    # specific errors
    if config.PRIMER_MAX_POLYX < 1:
        raise_error(
            "max polyx cannot be lower than 1.",
            log_file,
            exit=True
        )
    if config.PRIMER_MAX_DINUC_REPEATS < 1:
        raise_error(
            "max di-nucleotide repeats cannot be lower than 1.",
            log_file,
            exit=True
        )
    if config.BLAST_MAX_DIFF > 1:
        raise_error(
            "max difference to the blast db should be between 0-1.",
            log_file,
            exit=True
        )
    if config.BLAST_SIZE_MULTI < 1:
        raise_error(
            "off-targets should be considered at least in the range of the maximal amplicon length.",
            log_file,
            exit=True
        )
    if config.BLAST_PENALTY < 10:
        raise_error(
            "giving a too small penalty could result in the selection of off-target producing amplicons in the final scheme.",
            log_file,
        )
    # confirm proper BLAST settings in dictionary
    if not isinstance(config.BLAST_SETTINGS, dict):
        raise_error(
            "BLAST settings have to be in a dictionary format!",
            log_file,
            exit=True
        )
    if config.BLAST_SETTINGS["outfmt"] != "6 qseqid sseqid qlen length mismatch gapopen sstart send sstrand":
        raise_error(
            "output format (outfmt) of BLAST settings is not correct",
            log_file,
            exit=True
        )

    # check all dict values
    dict_values = ["evalue", "reward", "penalty", "gapopen", "gapextend"]

    for blast_setting in dict_values:
        if blast_setting not in config.BLAST_SETTINGS:
            raise_error(
                f"blast setting ({blast_setting}) is missing in BLAST dict",
                log_file,
                exit=True
            )
        if not isinstance(config.BLAST_SETTINGS[blast_setting], int):
            raise_error(
                f"blast setting ({blast_setting}) has to be an integer",
                log_file,
                exit=True
            )

    # write all settings to file
    var_dic = vars(config)
    with open(log_file, 'a') as f:
        print(
            "\nARG SETTINGS\n",
            f"THRESHOLD = {args.threshold}",
            f"PRIMER_ALLOWED_N_AMB = {args.n_ambig}",
            sep="\n",
            file=f
        )
        if args.mode in ("tiled", "single"):
            print(
                f"AMPLICON_OPT_LENGTH = {args.opt_length}",
                f"AMPLICON_MAX_LENGTH = {args.max_length}",
                sep="\n",
                file=f
            )
            if args.database is not None:
                print(
                    f"BLAST_DATABASE = {args.database}",
                    f"N_THREADS = {args.threads}",
                    sep="\n",
                    file=f
                )
        if args.mode == "tiled":
            print(
                f"MIN_OVERLAP = {args.overlap}",
                sep="\n",
                file=f
            )
        if args.mode == "single":
            print(
                f"REPORT_N_AMPLICONS = {args.report_n}",
                sep="\n",
                file=f
            )
        if args.mode == "qpcr":
            print(
                f"PROBE_ALLOWED_N_AMB = {args.pn_ambig}",
                f"TEST_DELTAG_N_AMPLICONS = {args.test_n}",
                f"DELTAG_CUTOFF = {args.deltaG}",
                sep="\n",
                file=f
            )
        print(
            "\nCONFIG SETTINGS\n",
            sep="\n",
            file=f
        )
        for var in all_vars[0]:
            print(f"{var} = {var_dic[var]}", file=f)
        if args.mode in ("tiled", "single"):
            if args.database is not None:
                for var in all_vars[2]:
                    print(f"{var} = {var_dic[var]}", file=f)
        if args.mode == "qpcr":
            for var in all_vars[1]:
                print(f"{var} = {var_dic[var]}", file=f)
        print("\nPROGRESS", file=f)


def goodbye_message():
    """
    This is what happens when I am bored...
    """

    messages = [
        "Thank you. Come again.",
        ">Placeholder for your advertisement<",
        "Make primers great again!",
        "Ciao cacao!"
        "And now lets pray to the PCR gods.",
        "**bibobibobop** task finished",
        "Thank you for traveling with varVAMP.",
        "This message will self-destruct in 0 seconds.... PUFF!",
        "What? Already finished? Errr I mean, thanks for using varVAMP!",
        "One primer design to rule them all.",
        "42",
        "Hope to see you again!",
        "Wohooooo everything worked somehow!",
        "Reminder: Get yourself a cup of coffee!",
        "Barba non facit philosophum.",
        "Task failed successfully.",
        "Never gonna give you up, never gonna let you down.",
    ]
    print(f"\n{random.choice(messages)}")



"""
amplicon search
"""

# BUILT-INS
import heapq
import math

# varVAMP
from varvamp.scripts import config, primers


def construct_graph(nodes, init_graph):
    """
    This method makes sure that the graph is symmetrical, but sets the costs
    for nodes in the reverse direction to infinity to make sure dijkstra
    never goes to an amplicon that is in the wrong direction.
    """
    graph = {}
    for node in nodes:
        graph[node] = {}

    graph.update(init_graph)

    for node, neighbors in graph.items():
        for neighbor in neighbors.keys():
            if graph[neighbor].get(node, False) is False:
                graph[neighbor][node] = float("infinity")

    return graph


class Graph(object):
    """
    a graph class
    """

    def __init__(self, nodes, init_graph):
        self.nodes = nodes
        self.graph = construct_graph(nodes, init_graph)

    def get_nodes(self):
        """
        Returns the nodes of the graph.
        """
        return self.nodes

    def get_neighbors(self, node):
        """
        Returns the neighbors of a node.
        """
        neighbors = []
        for out_node in self.nodes:
            if self.graph[node].get(out_node, False) is not False:
                neighbors.append(out_node)
        return neighbors

    def value(self, node1, node2):
        """
        Returns the value of an edge between two nodes.
        """
        return self.graph[node1][node2]


def find_amplicons(all_primers, opt_len, max_len):
    """
    finds all possible amplicons, creates a dictionary
    """
    amplicon_number = 0
    amplicon_dict = {}

    for left_name in all_primers["+"]:
        left_primer = all_primers["+"][left_name]
        for right_name in all_primers["-"]:
            right_primer = all_primers["-"][right_name]
            amplicon_length = right_primer[2] - left_primer[1]
            if not opt_len <= amplicon_length <= max_len:
                continue
            if primers.calc_dimer(left_primer[0], right_primer[0]).tm > config.PRIMER_MAX_DIMER_TMP:
                continue
            # calculate length dependend amplicon costs as the cumulative primer
            # penalty multiplied by the e^(fold length of the optimal length).
            amplicon_costs = (right_primer[3] + left_primer[3])*math.exp(amplicon_length/opt_len)
            amplicon_name = "AMPLICON_"+str(amplicon_number)
            amplicon_dict[amplicon_name] = [
                left_primer[1],  # start
                right_primer[2],  # stop
                left_name,  # name left primer
                right_name,  # name right primer
                amplicon_length,  # amplicon length
                amplicon_costs  # costs
            ]
            amplicon_number += 1

    return amplicon_dict


def create_amplicon_graph(amplicons, min_overlap):
    """
    creates the amplicon graph.
    """
    # ini graph and vertices
    amplicon_graph = {}
    nodes = []

    # add the maximum len of a primer to ensure that possible amplicon starts
    # before the min overlap
    min_overlap = min_overlap + config.PRIMER_SIZES[2]

    for current in amplicons:
        # remember all vertices
        nodes.append(current)
        current_amplicon = amplicons[current]
        start = current_amplicon[0] + current_amplicon[4]/2
        stop = current_amplicon[1] - min_overlap
        for next_amp in amplicons:
            next_amplicon = amplicons[next_amp]
            # check if the next amplicon lies within the start/stop range of
            # the current amplicon and if its non-overlapping part is large
            # enough to ensure space for a primer and the min overlap of the
            # following amplicon.
            if not all([start <= next_amplicon[0] <= stop, next_amplicon[1] > current_amplicon[1] + next_amplicon[4]/2]):
                continue
            if current not in amplicon_graph:
                amplicon_graph[current] = {next_amp: next_amplicon[5]}
            else:
                amplicon_graph[current][next_amp] = next_amplicon[5]

    # return a graph object
    return Graph(nodes, amplicon_graph)


def dijkstra_algorithm(graph, start_node):
    """
    implementation of the dijkstra algorithm
    """

    previous_nodes = {}
    shortest_path = {node: float('infinity') for node in graph.get_nodes()}
    shortest_path[start_node] = 0

    nodes_to_test = [(0, start_node)]

    while nodes_to_test:
        current_distance, current_node = heapq.heappop(nodes_to_test)
        if current_distance > shortest_path[current_node]:
            continue
        for neighbor in graph.get_neighbors(current_node):
            distance = current_distance + graph.value(current_node, neighbor)
            # Only consider this new path if it's a better path
            if not distance < shortest_path[neighbor]:
                continue
            shortest_path[neighbor] = distance
            previous_nodes[neighbor] = current_node
            heapq.heappush(nodes_to_test, (distance, neighbor))

    return previous_nodes, shortest_path


def get_end_node(previous_nodes, shortest_path, amplicons):
    """
    get the target node with the lowest penalty costs out of all
    nodes that have the same maximum end position
    """
    stop_nucleotide, possible_end_nodes = 0, {}

    for node in previous_nodes.keys():
        # check if a node has a larger stop -> empty dict and set new
        # best stop nucleotide
        if amplicons[node][1] > stop_nucleotide:
            possible_end_nodes = {node: shortest_path[node]}
            stop_nucleotide = amplicons[node][1]
        # if nodes have the same stop nucleotide, add to dictionary
        elif amplicons[node][1] == stop_nucleotide:
            possible_end_nodes[node] = shortest_path[node]

    # return the end node with the lowest penalty costs
    return min(possible_end_nodes.items(), key=lambda x: x[1])


def get_min_path(previous_nodes, start_node, target_node):
    """
    get the min path from the start to stop node from the
    previosuly calculated shortest path
    """
    path = []
    node = target_node

    while node != start_node:
        path.append(node)
        node = previous_nodes[node]
    # Add the start node manually
    path.append(start_node)

    # return the inverse list
    return path[::-1]


def create_scheme_dic(amplicon_scheme, amplicons, all_primers):
    """
    creates the final scheme dictionary
    """

    scheme_dictionary = {
        0: {},
        1: {}
    }

    for pool in (0, 1):
        for amp in amplicon_scheme[pool::2]:
            scheme_dictionary[pool][amp] = {}
            primer_pair = [amplicons[amp][2], amplicons[amp][3]]
            scheme_dictionary[pool][amp][primer_pair[0]] = all_primers["+"][primer_pair[0]]
            scheme_dictionary[pool][amp][primer_pair[1]] = all_primers["-"][primer_pair[1]]

    return scheme_dictionary


def find_best_covering_scheme(amplicons, amplicon_graph, all_primers):
    """
    this brute forces the amplicon scheme search until the largest
    coverage with the minimal costs is achieved.
    """
    # ini
    best_coverage = 0
    max_stop = max(amplicons.items(), key=lambda x: x[1])[1][1]
    lowest_costs = float('infinity')

    for start_node in amplicons:
        # if the currently best coverage + start nucleotide of the currently tested amplicon
        # is smaller than the maximal stop nucleotide there might be a better amplicon
        # scheme that covers more of the genome
        if amplicons[start_node][0] + best_coverage <= max_stop:
            previous_nodes, shortest_path = dijkstra_algorithm(amplicon_graph, start_node)
            # only continue if there are previous_nodes
            if previous_nodes:
                target_node, costs = get_end_node(previous_nodes, shortest_path, amplicons)
                coverage = amplicons[target_node][1] - amplicons[start_node][0]
                # if the new coverage is larger, go for the larger coverage
                if coverage > best_coverage:
                    best_start_node = start_node
                    best_target_node = target_node
                    best_previous_nodes = previous_nodes
                    lowest_costs = costs
                    best_coverage = coverage
                # if the coverages are identical, go for the lowest costs
                elif coverage == best_coverage:
                    if costs < lowest_costs:
                        best_start_node = start_node
                        best_target_node = target_node
                        best_previous_nodes = previous_nodes
                        lowest_costs = costs
                        best_coverage = coverage
            else:
                # check if the single amplicon has the largest coverage so far
                coverage = amplicons[start_node][1] - amplicons[start_node][0]
                if coverage > best_coverage:
                    best_start_node = start_node
                    best_previous_nodes = previous_nodes
                    lowest_costs = amplicons[start_node][5]
                    best_coverage = coverage
        # no need to check more, the best covering amplicon scheme was found and
        # has the minimal costs compared to the schemes with the same coverage
        else:
            break

    if best_previous_nodes:
        amplicon_scheme = get_min_path(best_previous_nodes, best_start_node, best_target_node)
    else:
        # if no previous nodes are found but the single amplicon results in the largest
        # coverage - return as the best scheme
        amplicon_scheme = [best_start_node]

    return best_coverage, create_scheme_dic(amplicon_scheme, amplicons, all_primers)


def test_scheme_for_dimers(amplicon_scheme):
    """
    test the lowest-cost scheme for primer dimers
    """

    primer_dimers = []

    for pool in amplicon_scheme:
        # test the primer dimers only within the respective pools
        tested_primers = []
        for amp in amplicon_scheme[pool]:
            for primer in amplicon_scheme[pool][amp]:
                # remember where the currrent primer was in the scheme
                current_primer = (pool, amp, primer, amplicon_scheme[pool][amp][primer])
                current_seq = current_primer[3][0]
                for tested in tested_primers:
                    tested_seq = tested[3][0]
                    if primers.calc_dimer(current_seq, tested_seq).tm <= config.PRIMER_MAX_DIMER_TMP:
                        continue
                    primer_dimers.append((current_primer, tested))
                # and remember all tested primers
                tested_primers.append(current_primer)

    return primer_dimers


def get_overlapping_primers(dimer, left_primer_candidates, right_primer_candidates):
    """
    get overlapping primers of a primer dimer pair that have been previously
    excluded. returns list of list with possible primers for each primer
    in primer dimer.
    """

    overlapping_primers = []
    # test each primer in dimer
    for primer in dimer:
        overlapping_primers_temp = []
        thirds_len = int((primer[3][2] - primer[3][1]) / 3)
        # get the middle third of the primer (here are the previously excluded primers)
        overlap_set = set(range(primer[3][1] + thirds_len, primer[3][2] - thirds_len + 1))
        # check in which list to look for them
        if "RIGHT" in primer[2]:
            primers_to_test = right_primer_candidates
        else:
            primers_to_test = left_primer_candidates
        # and check this list for all primers that overlap
        for potential_new in primers_to_test:
            primer_positions = list(range(potential_new[1], potential_new[2]+1))
            if not any(x in primer_positions for x in overlap_set):
                continue
            overlapping_primers_temp.append((primer[0], primer[1], primer[2], potential_new))

        overlapping_primers.append(overlapping_primers_temp)

    return overlapping_primers


def test_overlaps_for_dimers(overlapping_primers):
    """
    test the overlapping primers for dimers. return new primers.
    """
    for first_overlap in overlapping_primers[0]:
        for second_overlap in overlapping_primers[1]:
            # return the first match. primers are sorted by penalty.
            # first pair that makes it has the lowest penalty
            if primers.calc_dimer(first_overlap[3][0], second_overlap[3][0]).tm <= config.PRIMER_MAX_DIMER_TMP:
                return [first_overlap, second_overlap]


def check_and_solve_heterodimers(amplicon_scheme, left_primer_candidates, right_primer_candidates, all_primers):
    """
    check scheme for heterodimers, try to find
    new primers that overlap and replace the existing ones.
    this can lead to new primer dimers. therefore the scheme
    is checked a second time. if there are still primer dimers
    present the non-solvable dimers are returned
    """

    primer_dimers = test_scheme_for_dimers(amplicon_scheme)

    if primer_dimers:
        print(f"varVAMP found {len(primer_dimers)} dimer pairs in scheme ... trying to find replacements")
    else:
        return []

    for dimer in primer_dimers:
        # get overlapping primers that have not been considere
        overlapping_primers = get_overlapping_primers(dimer, left_primer_candidates, right_primer_candidates)
        # test all possible primers against each other for dimers
        new_primers = test_overlaps_for_dimers(overlapping_primers)
        # now change these primers in the scheme
        if new_primers:
            for new in new_primers:
                # overwrite in final scheme
                amplicon_scheme[new[0]][new[1]][new[2]] = new[3]
                # and in all primers
                if "LEFT" in new[2]:
                    strand = "+"
                else:
                    strand = "-"
                all_primers[strand][new[2]] = new[3]

    primer_dimers = test_scheme_for_dimers(amplicon_scheme)

    return primer_dimers


def find_single_amplicons(amplicons, all_primers, n):
    """
    find non-overlapping amplicons with low penalties
    from all found amplicons. only for the SINGLE mode.
    """
    # sort amplicons
    sorted_amplicons = sorted(amplicons.items(), key=lambda x: x[1][5])
    to_retain = [sorted_amplicons[0]]
    amplicon_range = list(range(sorted_amplicons[0][1][0], sorted_amplicons[0][1][1]+1))
    amplicon_set = set(amplicon_range)
    # find lowest non-overlapping
    for amp in sorted_amplicons:
        amp_positions = list(range(amp[1][0], amp[1][1]+1))
        if not any(x in amp_positions for x in amplicon_set):
            amplicon_set.update(amp_positions)
            to_retain.append(amp)
    # build the final dictionary
    scheme_dictionary = {0: {}}
    counter = 1
    for amp in to_retain:
        scheme_dictionary[0][amp[0]] = {}
        scheme_dictionary[0][amp[0]][amp[1][2]] = all_primers["+"][amp[1][2]]
        scheme_dictionary[0][amp[0]][amp[1][3]] = all_primers["-"][amp[1][3]]
        if n != float("inf"):
            if counter == n:
                break
            counter += 1

    return scheme_dictionary
